/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*      ------------         Bit-Bucket Software, Co.                       */
/*      \ 10001101 /         Writers and Distributors of                    */
/*       \ 011110 /          Freely Available<tm> Software.                 */
/*        \ 1011 /                                                          */
/*         ------                                                           */
/*                                                                          */
/*   (C) Copyright 1989, Bit Bucket Software Co., a Delaware Corporation.   */
/*                                                                          */
/*                                                                          */
/*                      Zmodem file reception module                        */
/*                                                                          */
/*                                                                          */
/*    For complete  details  of the licensing restrictions, please refer    */
/*    to the License  agreement,  which  is published in its entirety in    */
/*    the MAKEFILE and BT.C, and also contained in the file LICENSE.230.    */
/*                                                                          */
/*    USE  OF THIS FILE IS SUBJECT TO THE  RESTRICTIONS CONTAINED IN THE    */
/*    BINKLEYTERM  LICENSING  AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF    */
/*    THIS  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES,  OR IF YOU DO    */
/*    NOT HAVE THESE FILES,  YOU  SHOULD  IMMEDIATELY CONTACT BIT BUCKET    */
/*    SOFTWARE CO.  AT ONE OF THE  ADDRESSES  LISTED BELOW.  IN NO EVENT    */
/*    SHOULD YOU  PROCEED TO USE THIS FILE  WITHOUT HAVING  ACCEPTED THE    */
/*    TERMS  OF  THE  BINKLEYTERM  LICENSING  AGREEMENT,  OR  SUCH OTHER    */
/*    AGREEMENT AS YOU ARE ABLE TO REACH WITH BIT BUCKET SOFTWARE, CO.      */
/*                                                                          */
/*                                                                          */
/* You can contact Bit Bucket Software Co. at any one of the following      */
/* addresses:                                                               */
/*                                                                          */
/* Bit Bucket Software Co.        FidoNet  1:132/101, 1:132/491, 1:141/491  */
/* 427-3 Amherst Street           AlterNet 7:491/0                          */
/* CS2032, Suite 232              BBS-Net  86:2030/1                        */
/* Nashua, NH 03061               Internet f491.n132.z1.fidonet.org         */
/*                                                                          */
/* Please feel free to contact us at any time to share your comments about  */
/* our software and/or licensing policies.                                  */
/*                                                                          */
/*                                                                          */
/*  This module is based largely on a similar module in OPUS-CBCS V1.03b.   */
/*  The original work is (C) Copyright 1986, Wynn Wagner III. The original  */
/*  authors have graciously allowed us to use their code in this work.      */
/*                                                                          */
/*--------------------------------------------------------------------------*/


#include <time.h>
#include <string.h>
#include <conio.h>
#include <io.h>
#include <stdlib.h>

#ifdef __TURBOC__
#include "tc_utime.h"
#else
#include <sys/utime.h>
#endif

#include "com.h"
#include "xfer.h"
#include "zmodem.h"
#include "keybd.h"
#include "sbuf.h"
#include "sched.h"
#include "externs.h"
#include "prototyp.h"


/*--------------------------------------------------------------------------*/
/* Local routines                                                           */
static int RZ_ReceiveData (byte *, int);
static int RZ_32ReceiveData (byte *, int);
static int RZ_InitReceiver (void);
static int RZ_ReceiveBatch (FILE *);
static int RZ_ReceiveFile (FILE *);
static int RZ_GetHeader (void);
static int RZ_SaveToDisk (unsigned long *);
static void RZ_AckBibi (void);

/*--------------------------------------------------------------------------*/
/* Private declarations                                                     */
/*--------------------------------------------------------------------------*/
static long DiskAvail;
static long filetime;
static byte realname[64];

/*--------------------------------------------------------------------------*/
/* Private data                                                             */
/*--------------------------------------------------------------------------*/

/* Parameters for ZSINIT frame */
#define ZATTNLEN 32

static char Attn[ZATTNLEN + 1];                  /* String rx sends to tx on
                                                  * err            */
static FILE *Outfile;                            /* Handle of file being
                                                  * received           */
static int Tryzhdrtype;                          /* Hdr type to send for Last
                                                  * rx close      */
static char isBinary;                            /* Current file is binary
                                                  * mode             */
static char EOFseen;                             /* indicates cpm eof (^Z)
                                                  * was received     */
static char Zconv;                               /* ZMODEM file conversion
                                                  * request          */
static int RxCount;                              /* Count of data bytes
                                                  * received            */
static byte Upload_path[PATHLEN];                /* Dest. path of file being
                                                  * received   */
static long Filestart;                           /* File offset we started
                                                  * this xfer from   */
extern long filelength ();                       /* returns length of file
                                                  * ref'd in stream  */

/*--------------------------------------------------------------------------*/
/* GET ZMODEM                                                               */
/* Receive a batch of files.                                                */
/* returns TRUE (1) for good xfer, FALSE (0) for bad                        */
/* can be called from f_upload or to get mail from a WaZOO Opus             */
/*--------------------------------------------------------------------------*/
int get_Zmodem (rcvpath, xferinfo)
char *rcvpath;
FILE *xferinfo;
{
   byte namebuf[PATHLEN];
   int i;
   byte *p;
   long t, timerset ();

#ifdef DEBUG
   show_debug_name ("get_Zmodem");
#endif

   filetime = 0;
   errno = 0;

   _BRK_DISABLE ();
   IN_XON_ENABLE ();

/*   Secbuf = NULL;*/
   Outfile = NULL;
   z_size = 0;


   Rxtimeout = 100;
   Tryzhdrtype = ZRINIT;

   strcpy (namebuf, rcvpath);
   Filename = namebuf;

   strcpy (Upload_path, rcvpath);
   p = Upload_path + strlen (Upload_path) - 1;
   while (p >= Upload_path && *p != '\\')
      --p;
   *(++p) = '\0';
#ifdef BINKLEY  /*PLF Tue  10-31-1989  20:51:41 */
   HoldName = HoldAreaNameMunge(called_zone);

   sprintf (Abortlog_name, "%s%04x%04x.Z\0",
            HoldName, remote_net, remote_node);
#endif
   DiskAvail = zfree (Upload_path);
 
 #ifdef DEBUGLOG
   logit("@DiskAvail=%ld",DiskAvail);
 #endif

   if (((i = RZ_InitReceiver ()) == ZCOMPL) ||
       ((i == ZFILE) && ((RZ_ReceiveBatch (xferinfo)) == OK)))
      {
      XON_DISABLE ();
      XON_ENABLE ();                             /* Make sure xmitter is
                                                  * unstuck */
      return 1;
      }

   CLEAR_OUTBOUND ();
   XON_DISABLE ();                               /* Make sure xmitter is
                                                  * unstuck */
   send_can ();                                  /* transmit at least 10 cans    */
   t = timerset (200);                           /* wait no more than 2
                                                  * seconds  */
   while (!timeup (t) && !OUT_EMPTY () && CARRIER())
      time_release ();                           /* Give up slice while
                                                  * waiting  */
   XON_ENABLE ();                                /* Turn XON/XOFF back on...     */
/*
   if (Secbuf)
      free (Secbuf);
*/
   if (Outfile)
      fclose (Outfile);

   return 0;
}                                                /* get_Zmodem */

/*--------------------------------------------------------------------------*/
/* RZ RECEIVE DATA                                                          */
/* Receive array buf of max length with ending ZDLE sequence                */
/* and CRC.  Returns the ending character or error code.                    */
/*--------------------------------------------------------------------------*/
static int RZ_ReceiveData (buf, length)
REGISTER byte *buf;
REGISTER int length;
{
   REGISTER int c;
   REGISTER word crc;
   char *endpos;
   int d;


#ifdef DEBUG
   show_debug_name ("RZ_ReceiveData");
#endif

   if (Rxframeind == ZBIN32)
      return RZ_32ReceiveData (buf, length);

   crc = RxCount = 0;
   buf[0] = buf[1] = 0;
   endpos = buf + length;

   while (buf <= endpos)
      {
      if ((c = Z_GetZDL ()) & ~0xFF)
         {
   CRCfoo:
         switch (c)
            {
            case GOTCRCE:
            case GOTCRCG:
            case GOTCRCQ:
            case GOTCRCW:
               /*-----------------------------------*/
               /* C R C s                           */
               /*-----------------------------------*/
               crc = Z_UpdateCRC (((d = c) & 0xFF), crc);
               if ((c = Z_GetZDL ()) & ~0xFF)
                  goto CRCfoo;

               crc = Z_UpdateCRC (c, crc);
               if ((c = Z_GetZDL ()) & ~0xFF)
                  goto CRCfoo;

               crc = Z_UpdateCRC (c, crc);
               if (crc & 0xFFFF)
                  {
                  z_message (CRC_msg);
                  return ERROR;
                  }

               RxCount = length - (endpos - buf);
               return d;

            case GOTCAN:
               /*-----------------------------------*/
               /* Cancel                            */
               /*-----------------------------------*/
               z_log (Cancelled_msg);
               return ZCAN;

            case TIMEOUT:
               /*-----------------------------------*/
               /* Timeout                           */
               /*-----------------------------------*/
               z_message (TIME_msg);
               return c;

            case RCDO:
               /*-----------------------------------*/
               /* No carrier                        */
               /*-----------------------------------*/
               z_log (CARRIER_msg);
               CLEAR_INBOUND ();
               return c;

            default:
               /*-----------------------------------*/
               /* Something bizarre                 */
               /*-----------------------------------*/
               z_message ("Debris");
/*             cprintf (" [%x]", c);*/
               CLEAR_INBOUND ();
               return c;
            }                                    /* switch */
         }                                       /* if */

      *buf++ = (unsigned char) c;
      crc = Z_UpdateCRC (c, crc);
      }                                          /* while(1) */

   z_message ("Long pkt");
   return ERROR;
}                                                /* RZ_ReceiveData */

/*--------------------------------------------------------------------------*/
/* RZ RECEIVE DATA with 32 bit CRC                                          */
/* Receive array buf of max length with ending ZDLE sequence                */
/* and CRC.  Returns the ending character or error code.                    */
/*--------------------------------------------------------------------------*/
static int RZ_32ReceiveData (buf, length)
REGISTER byte *buf;
REGISTER int length;
{
   REGISTER int c;
   unsigned long crc;
   char *endpos;
   int d;


#ifdef DEBUG
   show_debug_name ("RZ_32ReceiveData");
#endif

   crc = 0xFFFFFFFF;
   RxCount = 0;
   buf[0] = buf[1] = 0;
   endpos = buf + length;

   while (buf <= endpos)
      {
      if ((c = Z_GetZDL ()) & ~0xFF)
         {
   CRCfoo:
         switch (c)
            {
            case GOTCRCE:
            case GOTCRCG:
            case GOTCRCQ:
            case GOTCRCW:
               /*-----------------------------------*/
               /* C R C s                           */
               /*-----------------------------------*/
               d = c;
               c &= 0377;
               crc = Z_32UpdateCRC (c, crc);
               if ((c = Z_GetZDL ()) & ~0xFF)
                  goto CRCfoo;

               crc = Z_32UpdateCRC (c, crc);
               if ((c = Z_GetZDL ()) & ~0xFF)
                  goto CRCfoo;

               crc = Z_32UpdateCRC (c, crc);
               if ((c = Z_GetZDL ()) & ~0xFF)
                  goto CRCfoo;

               crc = Z_32UpdateCRC (c, crc);
               if ((c = Z_GetZDL ()) & ~0xFF)
                  goto CRCfoo;

               crc = Z_32UpdateCRC (c, crc);
               if (crc != 0xDEBB20E3)
                  {
                  z_message (CRC_msg);
                  return ERROR;
                  }

               RxCount = length - (endpos - buf);
               return d;

            case GOTCAN:
               /*-----------------------------------*/
               /* Cancel                            */
               /*-----------------------------------*/
               z_log (Cancelled_msg);
               return ZCAN;

            case TIMEOUT:
               /*-----------------------------------*/
               /* Timeout                           */
               /*-----------------------------------*/
               z_message (TIME_msg);
               return c;

            case RCDO:
               /*-----------------------------------*/
               /* No carrier                        */
               /*-----------------------------------*/
               z_log (CARRIER_msg);
               CLEAR_INBOUND ();
               return c;

            default:
               /*-----------------------------------*/
               /* Something bizarre                 */
               /*-----------------------------------*/
               z_message ("Debris");
/*             cprintf (" [%x]", c);*/
               CLEAR_INBOUND ();
               return c;
            }                                    /* switch */
         }                                       /* if */

      *buf++ = (unsigned char) c;
      crc = Z_32UpdateCRC (c, crc);
      }                                          /* while(1) */

   z_message ("Long pkt");
   return ERROR;
}                                                /* RZ_ReceiveData */

/*--------------------------------------------------------------------------*/
/* RZ INIT RECEIVER                                                         */
/* Initialize for Zmodem receive attempt, try to activate Zmodem sender     */
/* Handles ZSINIT, ZFREECNT, and ZCOMMAND frames                            */
/*                                                                          */
/* Return codes:                                                            */
/*    ZFILE .... Zmodem filename received                                   */
/*    ZCOMPL ... transaction finished                                       */
/*    ERROR .... any other condition                                        */
/*--------------------------------------------------------------------------*/
static int RZ_InitReceiver ()
{
   REGISTER int n;
   int errors = 0;
   char *sptr;



#ifdef DEBUG
   show_debug_name ("RZ_InitReceiver");
#endif

   for (n = 12; --n >= 0;)
      {
      /*--------------------------------------------------------------*/
      /* Set buffer length (0=unlimited, don't wait).                 */
      /* Also set capability flags                                    */
      /*--------------------------------------------------------------*/
      Z_PutLongIntoHeader (0L);
      Txhdr[ZF0] = CANFC32 | CANFDX | CANOVIO;
      Z_SendHexHeader (Tryzhdrtype, Txhdr);
      if (Tryzhdrtype == ZSKIP)
         Tryzhdrtype = ZRINIT;

AGAIN:

      switch (Z_GetHeader (Rxhdr))
         {
         case ZFILE:
            Zconv = Rxhdr[ZF0];
            Tryzhdrtype = ZRINIT;
            if (RZ_ReceiveData (Secbuf, WAZOOMAX) == GOTCRCW)
               return ZFILE;
            Z_SendHexHeader (ZNAK, Txhdr);
            if (--n < 0)
               {
               sptr = "ZFILE";
               goto Err;
               }
            goto AGAIN;

         case ZSINIT:
            if (RZ_ReceiveData (Attn, ZATTNLEN) == GOTCRCW)
               {
               Z_PutLongIntoHeader (1L);
               Z_SendHexHeader (ZACK, Txhdr);
               }
            else Z_SendHexHeader (ZNAK, Txhdr);
            if (--n < 0)
               {
               sptr = "ZSINIT";
               goto Err;
               }
            goto AGAIN;

         case ZFREECNT:
            Z_PutLongIntoHeader (DiskAvail);
            Z_SendHexHeader (ZACK, Txhdr);
            goto AGAIN;

         case ZCOMMAND:
            /*-----------------------------------------*/
            /* Paranoia is good for you...             */
            /* Ignore command from remote, but lie and */
            /* say we did the command ok.              */
            /*-----------------------------------------*/
            if (RZ_ReceiveData (Secbuf, WAZOOMAX) == GOTCRCW)
               {
               status_line ("!Ignoring `%s'", Secbuf);
               Z_PutLongIntoHeader (0L);
               do
                  {
                  Z_SendHexHeader (ZCOMPL, Txhdr);
                  }
               while (++errors < 10 && Z_GetHeader (Rxhdr) != ZFIN);
               RZ_AckBibi ();
               return ZCOMPL;
               }
            else Z_SendHexHeader (ZNAK, Txhdr);
            if (--n < 0)
               {
               sptr = "CMD";
               goto Err;
               }
            goto AGAIN;

         case ZCOMPL:
            if (--n < 0)
               {
               sptr = "COMPL";
               goto Err;
               }
            goto AGAIN;

         case ZFIN:
            RZ_AckBibi ();
            return ZCOMPL;

         case ZCAN:
            sptr = CAN_msg;
            goto Err;

         case RCDO:
            sptr = CARRIER_msg;
            CLEAR_INBOUND ();
            goto Err;
         }                                       /* switch */
      }                                          /* for */

   sptr = TIME_msg;

Err:
   sprintf (e_input, "!Z-InitRecv %s", sptr);
   z_log (e_input);

   return ERROR;
}                                                /* RZ_InitReceiver */

/*--------------------------------------------------------------------------*/
/* RZFILES                                                                  */
/* Receive a batch of files using ZMODEM protocol                           */
/*--------------------------------------------------------------------------*/
static int RZ_ReceiveBatch (xferinfo)
FILE *xferinfo;
{
   REGISTER int c;


#ifdef DEBUG
   show_debug_name ("RZ_ReceiveBatch");
#endif

   while (1)
      {
      switch (c = RZ_ReceiveFile (xferinfo))
         {
         case ZEOF:
#ifdef BINKLEY              /*PLF Tue  10-31-1989  20:37:42 */
            if (Resume_WaZOO)
               {
               errno = 0;
               abortlog = fopen (Abortlog_name, read_ascii);
               if (!got_error (OPEN_msg, Abortlog_name))
                  {
                  strcpy (namebuf, Abortlog_name);
                  strcpy (namebuf + strlen (namebuf) - 1, "TMP");
                  errno = c = 0;
                  newlog = fopen (namebuf, write_ascii);
                  if (!got_error (OPEN_msg, namebuf))
                     {
                     while (!feof (abortlog))
                        {
                        linebuf[0] = '\0';
                        if (!fgets (linebuf, 64, abortlog))
                           break;
                        p = linebuf;
                        while (*p > ' ')
                           ++p;
                        *p = '\0';
                        if (stricmp (linebuf, Resume_name))
                           {
                           *p = ' ';
                           fputs (linebuf, newlog);
                           ++c;
                           }
                        }
                     fclose (abortlog);
                     fclose (newlog);
                     unlink (Abortlog_name);
                     if (c)
                        rename (namebuf, Abortlog_name);
                     else unlink (namebuf);
                     }
                  else fclose (abortlog);
                  }
               strcpy (namebuf, Upload_path);
               strcat (namebuf, Resume_name);
               unique_name (namebuf);
               rename (Filename, namebuf);
               }
#endif
            /* fallthrough */
         case ZSKIP:
            switch (RZ_InitReceiver ())
               {
               case ZCOMPL:
                  return OK;
               default:
                  return ERROR;
               case ZFILE:
                  break;
               }                                 /* switch */
            break;

         default:
            fclose (Outfile);
            Outfile = NULL;
#ifdef BINKLEY              /*PLF Tue  10-31-1989  20:25:11 */
            if (remote_capabilities)
               {
               if (!Resume_WaZOO)
                  {
                  strcpy (namebuf, Upload_path);
                  strcat (namebuf, "BadWaZOO.001");
                  unique_name (namebuf);
                  rename (Filename, namebuf);
                  errno = 0;
                  abortlog = fopen (Abortlog_name, "at");
                  if (!got_error (OPEN_msg, Abortlog_name))
                     {
                     fprintf (abortlog, "%s %s %s\n",
                              Resume_name,
                              namebuf + strlen (Upload_path),
                              Resume_info);
                     fclose (abortlog);
                     }
                  else unlink (namebuf);
                  }
               }
            else
#endif
            unlink (Filename);

            errno = 0;
            return c;
         }                                       /* switch */
      }                                          /* while */
}                                                /* RZ_ReceiveBatch */

/*--------------------------------------------------------------------------*/
/* RZ RECEIVE FILE                                                          */
/* Receive one file; assumes file name frame is preloaded in Secbuf         */
/*--------------------------------------------------------------------------*/
static int RZ_ReceiveFile (xferinfo)
FILE *xferinfo;
{
   REGISTER int c;
   int n;
   long rxbytes;
   char *sptr;
   struct utimbuf utimes;
   char j[50];


#ifdef DEBUG
   show_debug_name ("RZ_ReceiveFile");
#endif

   EOFseen = FALSE;
   c = RZ_GetHeader ();
   if (c == ERROR || c == ZSKIP)
      return (Tryzhdrtype = ZSKIP);

   n = 10;
   rxbytes = Filestart;

   while (1)
      {
      Z_PutLongIntoHeader (rxbytes);
      Z_SendHexHeader (ZRPOS, Txhdr);
NxtHdr:
      switch (c = Z_GetHeader (Rxhdr))
         {
         case ZDATA:
            /*-----------------------------------------*/
            /* Data Packet                             */
            /*-----------------------------------------*/
            if (Rxpos != rxbytes)
               {
               if (--n < 0)
                  {
                  sptr = FUBAR_msg;
                  goto Err;
                  }
               sprintf (j, "Bad pos; %ld/%ld", rxbytes, Rxpos);
               z_message (j);
               Z_PutString (Attn);
               continue;
               }
      MoreData:
            switch (c = RZ_ReceiveData (Secbuf, WAZOOMAX))
               {
               case ZCAN:
                  sptr = Cancelled_msg;
                  goto Err;

               case RCDO:
                  sptr = CARRIER_msg;
                  CLEAR_INBOUND ();
                  goto Err;

               case ERROR:
                  /*-----------------------*/
                  /* CRC error             */
                  /*-----------------------*/
                  if (--n < 0)
                     {
                     sptr = FUBAR_msg;
                     goto Err;
                     }
                  show_loc (rxbytes, n);
                  Z_PutString (Attn);
                  continue;

               case TIMEOUT:
                  if (--n < 0)
                     {
                     sptr = TIME_msg;
                     goto Err;
                     }
                  show_loc (rxbytes, n);
                  continue;

               case GOTCRCW:
                  /*---------------------*/
                  /* End of frame          */
                  /*-----------------------*/
                  n = 10;
                  if (RZ_SaveToDisk (&rxbytes) == ERROR)
                     return ERROR;
                  Z_PutLongIntoHeader (rxbytes);
                  Z_SendHexHeader (ZACK, Txhdr);
                  goto NxtHdr;

               case GOTCRCQ:
                  /*---------------------*/
                  /* Zack expected         */
                  /*-----------------------*/
                  n = 10;
                  if (RZ_SaveToDisk (&rxbytes) == ERROR)
                     return ERROR;
                  Z_PutLongIntoHeader (rxbytes);
                  Z_SendHexHeader (ZACK, Txhdr);
                  goto MoreData;

               case GOTCRCG:
                  /*---------------------*/
                  /* Non-stop              */
                  /*-----------------------*/
                  n = 10;
                  if (RZ_SaveToDisk (&rxbytes) == ERROR)
                     return ERROR;
                  goto MoreData;

               case GOTCRCE:
                  /*---------------------*/
                  /* Header to follow      */
                  /*-----------------------*/
                  n = 10;
                  if (RZ_SaveToDisk (&rxbytes) == ERROR)
                     return ERROR;
                  goto NxtHdr;
               }                                 /* switch */

         case ZNAK:
         case TIMEOUT:
            /*-----------------------------------------*/
            /* Packet was probably garbled             */
            /*-----------------------------------------*/
            if (--n < 0)
               {
               sptr = "Garbled packet";
               goto Err;
               }
            show_loc (rxbytes, n);
            continue;

         case ZFILE:
            /*-----------------------------------------*/
            /* Sender didn't see our ZRPOS yet         */
            /*-----------------------------------------*/
            RZ_ReceiveData (Secbuf, WAZOOMAX);
            continue;

         case ZEOF:
            /*-----------------------------------------*/
            /* End of the file                         */
            /* Ignore EOF if it's at wrong place; force */
            /* a timeout because the eof might have    */
            /* gone out before we sent our ZRPOS       */
            /*-----------------------------------------*/
            if (Rxpos != rxbytes)
               goto NxtHdr;

            throughput (2, rxbytes - Filestart);
            errno = 0;

            if(fclose (Outfile))
                got_error (CLOSE_msg, Filename);

/*          status_line ("+Received-Z%s %s", Crc32 ? "/32" : "", realname);*/
            status_line ("=UL-Z %s", realname);

            update_files (0);

            if (filetime)
               {
               utimes.actime = filetime;
               utimes.modtime = filetime;
               utime (Filename, &utimes);
               }

            Outfile = NULL;
            if (xferinfo != NULL)
               {
               fprintf (xferinfo, "%s\n", Filename);
               got_error (WRITE_msg, "XferInfo");
               }
            return c;

         case ERROR:
            /*-----------------------------------------*/
            /* Too much garbage in header search error */
            /*-----------------------------------------*/
            if (--n < 0)
               {
               sptr = "HdrJunk";
               goto Err;
               }
            show_loc (rxbytes, n);
            Z_PutString (Attn);
            continue;

         case ZSKIP:
            return c;

         default:
            sptr = IDUNNO_msg;
            CLEAR_INBOUND ();
            goto Err;
         }                                       /* switch */
      }                                          /* while */

Err:sprintf (e_input, "!Z-rz %s", sptr);
   z_log (e_input);
   return ERROR;
}                                                /* RZ_ReceiveFile */

/*--------------------------------------------------------------------------*/
/* RZ GET HEADER                                                            */
/* Process incoming file information header                                 */
/*--------------------------------------------------------------------------*/
static int RZ_GetHeader ()
{
   REGISTER byte *p;
   struct stat f;
   int i;
   byte *ourname;
   byte *theirname;
   unsigned long filesize;
   byte *fileinfo;
   char j[80];


#ifdef DEBUG
   show_debug_name ("RZ_GetHeader");
#endif

   /*--------------------------------------------------------------------*/
   /* Setup the transfer mode                                            */
   /*--------------------------------------------------------------------*/
   isBinary = (char) ((!RXBINARY && Zconv == ZCNL) ? 0 : 1);


   /*--------------------------------------------------------------------*/
   /* Extract and verify filesize, if given.                             */
   /* Reject file if not at least 10K free                               */
   /*--------------------------------------------------------------------*/
   filesize = 0L;
   filetime = 0L;
   fileinfo = Secbuf + 1 + strlen (Secbuf);
   if (*fileinfo)
      sscanf (fileinfo, "%ld %lo", &filesize, &filetime);
    
  #ifdef DEBUGLOG
   logit("@diskavl=%ld",DiskAvail);
  #endif
    
   if (filesize + 10240 > DiskAvail)
      {
      status_line ("!Disk space");
      ul_no_space=TRUE;
      return ERROR;
      }

    #ifdef DEBUGLOG
      logit("@prm.k_free=%d, DiskAvail=%ld, FileSize=%ld",prm.k_free,
            DiskAvail,filesize);
      logit("@DiskAvail-filesize-10240L=%ld, prm.k_free*1000L=%ld",
            DiskAvail-filesize-10240L,(long)prm.k_free*1000L);
    #endif
      
    if (prm.k_free && DiskAvail-filesize-10240L < (long)prm.k_free*1000L)
    {
      logit("!Not enough free space for UL");
      ul_no_space=TRUE;
      return ERROR;
    }


   /*--------------------------------------------------------------------*/
   /* Get and/or fix filename for uploaded file                          */
   /*--------------------------------------------------------------------*/
   p = Filename + strlen (Filename) - 1;         /* Find end of upload path */
   while (p >= Filename && *p != '\\')
      p--;
   ourname = ++p;

   p = Secbuf + strlen (Secbuf) - 1;             /* Find transmitted simple
                                                  * filename */
   while (p >= Secbuf && *p != '\\' && *p != '/' && *p != ':')
      p--;
   theirname = ++p;

   strcpy (ourname, theirname);                  /* Start w/ our path & their
                                                  * name */
   strcpy (realname, Filename);

   /*--------------------------------------------------------------------*/
   /* Save info on WaZOO transfer in case of abort                       */
   /*--------------------------------------------------------------------*/
#ifdef BINKLEY      /*PLF Tue  10-31-1989  20:40:22 */
   Resume_WaZOO = 0;
   if (remote_capabilities)
      {
      strcpy (Resume_name, theirname);
      sprintf (Resume_info, "%ld %lo", filesize, filetime);
      }
   /*--------------------------------------------------------------------*/
   /* Check if this is a failed WaZOO transfer which should be resumed   */
   /*--------------------------------------------------------------------*/
   if (remote_capabilities && dexists (Abortlog_name))
      {
      errno = 0;
      abortlog = fopen (Abortlog_name, read_ascii);
      if (!got_error (OPEN_msg, Abortlog_name))
         {
         while (!feof (abortlog))
            {
            linebuf[0] = '\0';
            if (!fgets ((p = linebuf), 64, abortlog))
               break;
            while (*p >= ' ')
               ++p;
            *p = '\0';
            p = strchr (linebuf, ' ');
            *p = '\0';
            if (!stricmp (linebuf, theirname))
               {
               p = strchr ((badfile_name = ++p), ' ');
               *p = '\0';
               if (!stricmp (++p, Resume_info))
                  {
                  Resume_WaZOO = 1;
                  break;
                  }
               }
            }
         fclose (abortlog);
         }
      }
   /*--------------------------------------------------------------------*/
   /* Open either the old or a new file, as appropriate                  */
   /*--------------------------------------------------------------------*/
   if (Resume_WaZOO)
      {
      strcpy (ourname, badfile_name);
      if (dexists (Filename))
         p = "r+b";
      else p = write_binary;
      }
   else
#endif /* BINKLEY */
      {
      strcpy (ourname, theirname);
      /*--------------------------------------------------------------------*/
      /* If the file already exists:                                        */
      /* 1) And the new file has the same time and size, return ZSKIP    */
      /* 2) And OVERWRITE is turned on, delete the old copy              */
      /* 3) Else create a unique file name in which to store new data    */
      /*--------------------------------------------------------------------*/
      if (dexists (Filename))
         {                                       /* If file already exists...      */
         errno = 0;
         Outfile = fopen (Filename, read_binary);
         if (!Outfile){
            got_error (OPEN_msg, Filename);
            return ERROR;
        }
         fstat (fileno (Outfile), &f);
         if(fclose (Outfile))
             got_error (CLOSE_msg, Filename);
         if (filesize == f.st_size /* && filetime == f.st_mtime*/ )
            {
            status_line ("+Already have %s", Filename);
            return ZSKIP;
            }
         i = strlen (Filename) - 1;
         if ((!overwrite) || (is_arcmail (Filename, i)))
            {
            unique_name (Filename);
            }
         else
            {
            unlink (Filename);
            }
         }                                       /* if exist */

      if (strcmp (ourname, theirname))
         {
         status_line (RENAME_msg, ourname);
         }
      p = write_binary;
      }
   errno = 0;
   Outfile = fopen (Filename, p);
   if (!Outfile){
      got_error (OPEN_msg, Filename);
      return ERROR;
    }
   if (isatty (fileno (Outfile)))
      {
      errno = 1;
      got_error (DEVICE_msg, Filename);
      fclose (Outfile);
      return (ERROR);
      }


#ifdef BINKLEY
   Filestart = (Resume_WaZOO) ? filelength (fileno (Outfile)) : 0L;
   if (Resume_WaZOO)
      status_line (":Resuming from offset %ld", Filestart);
   fseek (Outfile, Filestart, SEEK_SET);

   if (remote_capabilities)
        p = check_netfile(theirname);
   else p = NULL;
#else
    Filestart = 0L;
    fseek (Outfile, 0L, SEEK_SET);
    p = NULL;
#endif

   sprintf (j, "%s %s; %s%ldb, %d min.\n",
            (p) ? p : "Receiving",
            realname,
            (isBinary) ? "" : "ASCII ",
            filesize,
            (int) (filesize * 10 / cur_baud + 27) / 54);

   Lputs(j);
   throughput (0, 0L);

   return OK;
}                                                /* RZ_GetHeader */

/*--------------------------------------------------------------------------*/
/* RZ SAVE TO DISK                                                          */
/* Writes the received file data to the output file.                        */
/* If in ASCII mode, stops writing at first ^Z, and converts all            */
/*   solo CR's or LF's to CR/LF pairs.                                      */
/*--------------------------------------------------------------------------*/
static int RZ_SaveToDisk (rxbytes)
unsigned long *rxbytes;
{
   static byte lastsent;

   REGISTER byte *p;
   REGISTER int count;
   int i;

#ifdef DEBUG
   show_debug_name ("RZ_SaveToDisk");
#endif

   count = RxCount;

   if (got_ESC ())
      {
      send_can ();                               /* Cancel file */
      while ((i = Z_GetByte (20)) != TIMEOUT && i != RCDO)      /* Wait for line to
                                                                 * clear */
         CLEAR_INBOUND ();
      send_can ();                               /* and Cancel Batch */
      z_log (KBD_msg);
      return ERROR;
      }

   if (count != z_size)
      {
      z_size = count;
      Lprintf("          %u\r", count);
      }

   errno = 0;

   if (isBinary)
      {
      fwrite (Secbuf, count, 1, Outfile);
      if (errno)
         goto oops;
      }
   else
      {
      if (EOFseen)
         return OK;
      for (p = Secbuf; --count >= 0; ++p)
         {
         if (*p == CPMEOF)
            {
            EOFseen = TRUE;
            return OK;
            }
         if (*p == '\n')
            {
            if (lastsent != '\r' && putc ('\r', Outfile) == EOF)
               goto oops;
            }
         else
            {
            if (lastsent == '\r' && putc ('\n', Outfile) == EOF)
               goto oops;
            }
         if (putc ((lastsent = *p), Outfile) == EOF)
            goto oops;
         }
      }

   *rxbytes += RxCount;
   Lprintf("%lu\r", *rxbytes);
   return OK;

oops:
   got_error (WRITE_msg, Filename);
   return ERROR;

}                                                /* RZ_SaveToDisk */

/*--------------------------------------------------------------------------*/
/* RZ ACK BIBI                                                              */
/* Ack a ZFIN packet, let byegones be byegones                              */
/*--------------------------------------------------------------------------*/
static void RZ_AckBibi ()
{
   REGISTER int n;


#ifdef DEBUG
   show_debug_name ("RZ_AckBiBi");
#endif

   Z_PutLongIntoHeader (0L);
   for (n = 4; --n;)
      {
      Z_SendHexHeader (ZFIN, Txhdr);
      switch (Z_GetByte (100))
         {
         case 'O':
            Z_GetByte (1);                       /* Discard 2nd 'O' */

         case TIMEOUT:
         case RCDO:
            return;
         }                                       /* switch */
      }                                          /* for */
}                                                /* RZ_AckBibi */
